/*********************************************************/
/* TAUCS                                                 */
/* Author: Sivan Toledo                                  */
/*********************************************************/

#define TAUCS_INT     1024  /* default is double */

#define TAUCS_LOWER      1
#define TAUCS_UPPER      2
#define TAUCS_TRIANGULAR 4
#define TAUCS_SYMMETRIC  8
#define TAUCS_PATTERN    8

typedef struct
{
  int     n;    /* columns                      */
  int     m;    /* rows; don't use if symmetric   */
  int     flags;
  int*    colptr; /* pointers to where columns begin in rowind and values. */
                  /* 0-based. Length is (n+1). */
  int*    rowind; /* row indices */
  double* values; /* numerical values */
} taucs_ccs_matrix;

taucs_ccs_matrix* taucs_ccs_create(int m, int n, int nnz);
void              taucs_ccs_free  (taucs_ccs_matrix* matrix);
int               taucs_ccs_write_ijv(taucs_ccs_matrix* matrix, 
				      char* ijvfilename);
taucs_ccs_matrix* taucs_ccs_read_ijv (char* filename,int flags);
taucs_ccs_matrix* taucs_ccs_read_mtx (char* filename,int flags);
taucs_ccs_matrix* taucs_ccs_read_ccs (char* filename,int flags);
void              taucs_ccs_order(taucs_ccs_matrix* matrix, 
				  int** perm, int** invperm,
				  char* which);
taucs_ccs_matrix* taucs_ccs_generate_mesh2d(int n,char *which);
taucs_ccs_matrix* taucs_ccs_generate_mesh2d_negative(int n);
taucs_ccs_matrix* taucs_ccs_generate_mesh3d(int X, int Y, int Z);
taucs_ccs_matrix* taucs_ccs_generate_dense(int m,int n, int flags);
taucs_ccs_matrix* taucs_ccs_generate_rrn(int X, int Y, int Z, 
					 double drop_probability, 
					 double rmin);
taucs_ccs_matrix* taucs_ccs_generate_discontinuous(int X, int Y, int Z, 
						   double jump);
double* taucs_vec_generate_continuous(int X, int Y, int Z, char* which);

void              taucs_ccs_split(taucs_ccs_matrix* A, 
				  taucs_ccs_matrix** L, 
				  taucs_ccs_matrix** R, 
				  int p);
taucs_ccs_matrix* taucs_ccs_permute_symmetrically(taucs_ccs_matrix* A, 
						  int* perm, int* invperm);
void              taucs_ccs_times_vec(taucs_ccs_matrix* m, 
				      double* X,
				      double* B);

taucs_ccs_matrix* taucs_ccs_augment_nonpositive_offdiagonals(taucs_ccs_matrix* A);

taucs_ccs_matrix* taucs_ccs_factor_llt        (taucs_ccs_matrix* A,
					       double droptol, int modified);
taucs_ccs_matrix* taucs_ccs_factor_llt_partial(taucs_ccs_matrix* A, 
					       int p);
int               taucs_ccs_solve_llt  (void* L, double* x, double* b);
int               taucs_ccs_solve_schur(taucs_ccs_matrix* L,
					taucs_ccs_matrix* schur_comp,
					int    (*schur_precond_fn)(void*,double* x,double* b),
					void*  schur_precond_args,
					int    maxits,
					double convratio,
					double* x, double* b);

taucs_ccs_matrix* taucs_ccs_factor_ldlt(taucs_ccs_matrix* A);
int               taucs_ccs_solve_ldlt (void* L, double* x, double* b);

taucs_ccs_matrix* taucs_ccs_factor_xxt(taucs_ccs_matrix* A);
int               taucs_ccs_solve_xxt (void* X, double* x, double* b);

int taucs_conjugate_gradients(taucs_ccs_matrix*  A,
			      int               (*precond_fn)(void*,double* x,double* b),
			      void*             precond_args,
			      double*           X,
			      double*           B,
			      int               itermax,
			      double            convergetol);
int taucs_minres(taucs_ccs_matrix*  A,
		 int               (*precond_fn)(void*,double* x,double* b),
		 void*             precond_args,
		 double*           X,
		 double*           B,
		 int               itermax,
		 double            convergetol);

int
taucs_sg_preconditioner_solve(void*   P,
			      double* z, 
			      double* r);
void 
*taucs_sg_preconditioner_create(taucs_ccs_matrix *A,
				int **perm,
				int **invperm,
				char* ordering,
				char *specification);
void taucs_sg_preconditioner_free(void* P);

taucs_ccs_matrix*
taucs_amwb_preconditioner_create(taucs_ccs_matrix *symccs_mtxA, 
				int rnd,
				double subgraphs);

void* 
taucs_recursive_amwb_preconditioner_create(taucs_ccs_matrix* A, 
					   double c, 
					   double epsilon, 
					   int nsmall,
					   int maxlevels,
					   int innerits,
					   double convratio,
					   int** perm, 
					   int** invperm);

int
taucs_recursive_amwb_preconditioner_solve(void* P, 
					 double* Z, 
					 double* R);

int taucs_ccs_etree(taucs_ccs_matrix* A,
		    int* parent,
		    int* l_colcount,
		    int* l_rowcount,
		    int* l_nnz);

int      
taucs_ccs_symbolic_elimination(taucs_ccs_matrix* A,
			       void* L,
			       int do_order
			       );

void* taucs_ccs_factor_llt_symbolic(taucs_ccs_matrix* A);
int   taucs_ccs_factor_llt_numeric (taucs_ccs_matrix* A,void* L);

void* taucs_ccs_factor_llt_mf(taucs_ccs_matrix* A);
void* taucs_ccs_factor_llt_ll(taucs_ccs_matrix* A);
int   taucs_supernodal_solve_llt(void* vL,
	  		         double* x, double* b);
void taucs_supernodal_factor_free(void* L);
void taucs_supernodal_factor_free_numeric(void* L);
taucs_ccs_matrix*
     taucs_supernodal_factor_to_ccs(void* L);

void taucs_vec_permute(int n, double v[], double pv[], int p[]);
void taucs_vec_ipermute(int n, double pv[], double v[], int invp[]);

/*********************************************************/
/* Utilities                                             */
/*********************************************************/

void   taucs_logfile(char* file_prefix);
int    taucs_printf(char *fmt, ...);
double taucs_system_memory_size();
double taucs_available_memory_size();
double taucs_wtime();
double taucs_ctime();

/*********************************************************/
/* Out-of-core IO routines                               */
/*********************************************************/

typedef struct {
  int   type;
  int   nmatrices;
  void* type_specific;

  /* the following may change! do not rely on them. */
  double nreads, nwrites, bytes_read, bytes_written, read_time, write_time;
} taucs_io_handle;

taucs_io_handle* taucs_io_create_singlefile(char* filename);
taucs_io_handle* taucs_io_open_singlefile(char* filename);

taucs_io_handle* taucs_io_create_multifile(char* filename);
taucs_io_handle* taucs_io_open_multifile(char* filename);

int              taucs_io_close (taucs_io_handle* f);
int              taucs_io_delete(taucs_io_handle* f);

int              taucs_io_append(taucs_io_handle* f,
				 int   index,
				 int   m,int n,
				 int   flags,
				 void* data
				 );
int              taucs_io_write(taucs_io_handle* f,
				int   index,
				int   m,int n,
				int   flags,
				void* data
				);
int              taucs_io_read(taucs_io_handle* f,
			       int   index,
			       int   m,int n,
			       int   flags,
			       void* data
			       );

/*********************************************************/
/* Out-of-core Sparse Choleksy routines                  */
/*********************************************************/

int taucs_ooc_factor_llt(taucs_ccs_matrix* A, 
			 taucs_io_handle*  L,
			 double memory);
int taucs_ooc_solve_llt (void* L /* actual type: taucs_io_handle* */,
			 double* x, double* b);

/*********************************************************/
/*                                                       */
/*********************************************************/

#ifdef OSTYPE_irix
#include <ieeefp.h>

#define isinf(x) (!finite((x)) && !isnan((x)))
#endif

/*********************************************************/
/*                                                       */
/*********************************************************/




