/*********************************************************/
/* TAUCS                                                 */
/* Author: Sivan Toledo                                  */
/*********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include "taucs.h"

#define max(x,y) ( ((x) > (y)) ? (x) : (y) )

/*********************************************************/
/*                                                       */
/*********************************************************/
#if 0
/* split into left p columns, right p columns */
void 
taucs_symccs_split(symccs_matrix* A, 
		   symccs_matrix** L, 
		   symccs_matrix** R, 
		   int p)
{
  int i,n;
  int Lnnz, Rnnz;

  assert(A->uplo == 'l');

  n = A->n;

  *L = (symccs_matrix*) malloc(sizeof(symccs_matrix));
  *R = (symccs_matrix*) malloc(sizeof(symccs_matrix));
  if (!(*L) || !(*R)) { 
    	taucs_printf("symccs_split: out of memory (1)\n");
	free(*L);
	free(*R);
	*L = *R = NULL;
	return; 
  }

  Lnnz = 0;
  for (i=0; i<p; i++)
    Lnnz += ( (A->colptr)[i+1] - (A->colptr)[i] );
    
  (*L)->uplo = 'l';
  /*(*L)->indshift = A->indshift;*/
  (*L)->n = n;
  (*L)->colptr = (int*)    malloc((n+1) * sizeof(int));
  (*L)->rowind = (int*)    malloc(Lnnz   * sizeof(int));
  (*L)->values = (double*) malloc(Lnnz   * sizeof(double));
  if (!((*L)->colptr) || !((*L)->rowind) || !((*L)->rowind)) {
    	taucs_printf("symccs_split: out of memory (2): n=%d nnz=%d\n",n,Lnnz);
	free((*L)->colptr); free((*L)->rowind); free((*L)->values);
	free ((*L));
	return; 
  }

  for (i=0; i<=p; i++)
    ((*L)->colptr)[i] = (A->colptr)[i];   
  for (i=p+1; i<n+1; i++)
    ((*L)->colptr)[i] = ((*L)->colptr)[p]; /* other columns are empty */

  for (i=0; i<Lnnz; i++) {
    ((*L)->rowind)[i] = (A->rowind)[i];
    ((*L)->values)[i] = (A->values)[i];
  }

  /* now copy right part of matrix into a p-by-p matrix */

  Rnnz = 0;
  for (i=p; i<n; i++)
    Rnnz += ( (A->colptr)[i+1] - (A->colptr)[i] );
    
  (*R)->uplo = 'l';
  /*(*R)->indshift = A->indshift;*/
  (*R)->n = n-p;
  (*R)->colptr = (int*)    malloc((n-p+1) * sizeof(int));
  (*R)->rowind = (int*)    malloc(Rnnz   * sizeof(int));
  (*R)->values = (double*) malloc(Rnnz   * sizeof(double));
  if (!((*R)->colptr) || !((*R)->rowind) || !((*R)->rowind)) {
    	taucs_printf("symccs_split: out of memory (3): p=%d nnz=%d\n",p,Rnnz);
	free((*R)->colptr); free((*R)->rowind); free((*R)->values);
	free((*L)->colptr); free((*L)->rowind); free((*L)->values);
	free ((*R));
	free ((*L));
	return; 
  }
    
  for (i=0; i<=(n-p); i++)
    ((*R)->colptr)[i] = (A->colptr)[i+p] - Lnnz;   

  for (i=0; i<Rnnz; i++) {
    ((*R)->rowind)[i] = (A->rowind)[i + Lnnz] - p;
    ((*R)->values)[i] = (A->values)[i + Lnnz];
  }
} 

symccs_matrix* 
taucs_symccs_permute(symccs_matrix* A, int* perm, int* invperm)
{
  symccs_matrix* PAPT;
  int n;
  int nnz;
  int* colptr;
  int* len;
  int i,j,ip,I,J;
  double AIJ;

  n   = A->n;
  nnz = (A->colptr)[n];

  PAPT = taucs_symccs_create(n,nnz);
  PAPT->uplo = 'l';

  len    = (int*) malloc(n * sizeof(int));
  colptr = (int*) malloc(n * sizeof(int));

  for (j=0; j<n; j++) len[j] = 0;

  for (j=0; j<n; j++) {
    for (ip = (A->colptr)[j]; ip < (A->colptr)[j+1]; ip++) {
      /*i = (A->rowind)[ip] - (A->indshift);*/
      i = (A->rowind)[ip];

      I = invperm[i];
      J = invperm[j];

      if (I < J) {
	int T = I; 
	I = J;
	J = T;
      }

      len[J] ++;
      
    }
  }

  (PAPT->colptr)[0] = 0;
  for (j=1; j<=n; j++) (PAPT->colptr)[j] = (PAPT->colptr)[j-1] + len[j-1];

  for (j=0; j<n; j++) len[j] = (PAPT->colptr)[j];
  
  for (j=0; j<n; j++) {
    for (ip = (A->colptr)[j]; ip < (A->colptr)[j+1]; ip++) {
      /*i   = (A->rowind)[ip] - (A->indshift);*/
      i   = (A->rowind)[ip];
      AIJ = (A->values)[ip];

      I = invperm[i];
      J = invperm[j];

      if (I < J) {
	int T = I; 
	I = J;
	J = T;
      }

      /*(PAPT->rowind)[ len[J] ] = I + (PAPT->indshift);*/
      (PAPT->rowind)[ len[J] ] = I;
      (PAPT->values)[ len[J] ] = AIJ;

      len[J] ++;
    }
  }
  
  return PAPT;
}

/* compute B=A*X, A must be square */

void 
taucs_symccs_times_vec(symccs_matrix* m, 
		       double* X,
		       double* B)
{
  int i,ip,j,n;
  double Aij;

  n = m->n;
  
  for (i=0; i < n; i++) B[i] = 0.0;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr[j+1]); ip++) {
      /*i   = (m->rowind)[ip] - (m->indshift);*/
      i   = (m->rowind)[ip];
      Aij = (m->values)[ip];

      B[i] += X[j] * Aij;
      if (i != j) 
	B[j] += X[i] *Aij;
    }
  }
} 
#endif
/*********************************************************/
/* CCS                                                   */
/*********************************************************/

/* split into left p columns, right p columns */
void 
taucs_ccs_split(taucs_ccs_matrix*  A, 
		taucs_ccs_matrix** L, 
		taucs_ccs_matrix** R, 
		int p)
{
  int i,n;
  int Lnnz, Rnnz;

  assert((A->flags & TAUCS_SYMMETRIC) || (A->flags & TAUCS_TRIANGULAR));
  assert(A->flags & TAUCS_LOWER);

  n = A->n;

  *L = (taucs_ccs_matrix*) malloc(sizeof(taucs_ccs_matrix));
  *R = (taucs_ccs_matrix*) malloc(sizeof(taucs_ccs_matrix));
  if (!(*L) || !(*R)) { 
    taucs_printf("taucs_ccs_split: out of memory\n");
    free(*L);
    free(*R);
    *L = *R = NULL;
    return; 
  }

  Lnnz = 0;
  for (i=0; i<p; i++)
    Lnnz += ( (A->colptr)[i+1] - (A->colptr)[i] );
    
  (*L)->flags = TAUCS_SYMMETRIC | TAUCS_LOWER;
  (*L)->n = n;
  (*L)->m = n;
  (*L)->colptr = (int*)    malloc((n+1) * sizeof(int));
  (*L)->rowind = (int*)    malloc(Lnnz   * sizeof(int));
  (*L)->values = (double*) malloc(Lnnz   * sizeof(double));
  if (!((*L)->colptr) || !((*L)->rowind) || !((*L)->rowind)) {
    	taucs_printf("taucs_ccs_split: out of memory: n=%d nnz=%d\n",n,Lnnz);
	free((*L)->colptr); free((*L)->rowind); free((*L)->values);
	free ((*L));
	return; 
  }

  for (i=0; i<=p; i++)
    ((*L)->colptr)[i] = (A->colptr)[i];   
  for (i=p+1; i<n+1; i++)
    ((*L)->colptr)[i] = ((*L)->colptr)[p]; /* other columns are empty */

  for (i=0; i<Lnnz; i++) {
    ((*L)->rowind)[i] = (A->rowind)[i];
    ((*L)->values)[i] = (A->values)[i];
  }

  /* now copy right part of matrix into a p-by-p matrix */

  Rnnz = 0;
  for (i=p; i<n; i++)
    Rnnz += ( (A->colptr)[i+1] - (A->colptr)[i] );
    
  (*R)->flags = TAUCS_SYMMETRIC | TAUCS_LOWER;
  (*R)->n = n-p;
  (*R)->m = n-p;
  (*R)->colptr = (int*)    malloc((n-p+1) * sizeof(int));
  (*R)->rowind = (int*)    malloc(Rnnz   * sizeof(int));
  (*R)->values = (double*) malloc(Rnnz   * sizeof(double));
  if (!((*R)->colptr) || !((*R)->rowind) || !((*R)->rowind)) {
    	taucs_printf("taucs_ccs_split: out of memory (3): p=%d nnz=%d\n",p,Rnnz);
	free((*R)->colptr); free((*R)->rowind); free((*R)->values);
	free((*L)->colptr); free((*L)->rowind); free((*L)->values);
	free ((*R));
	free ((*L));
	return; 
  }
    
  for (i=0; i<=(n-p); i++)
    ((*R)->colptr)[i] = (A->colptr)[i+p] - Lnnz;   

  for (i=0; i<Rnnz; i++) {
    ((*R)->rowind)[i] = (A->rowind)[i + Lnnz] - p;
    ((*R)->values)[i] = (A->values)[i + Lnnz];
  }
} 

taucs_ccs_matrix* 
taucs_ccs_permute_symmetrically(taucs_ccs_matrix* A, int* perm, int* invperm)
{
  taucs_ccs_matrix* PAPT;
  int n;
  int nnz;
  /*int* colptr;*/
  int* len;
  int i,j,ip,I,J;
  double AIJ;

  /*assert(A->flags & TAUCS_SYMMETRIC);*/
  assert(A->flags & TAUCS_LOWER);

  n   = A->n;
  nnz = (A->colptr)[n];

  PAPT = taucs_ccs_create(n,n,nnz);
  /*PAPT->flags = TAUCS_SYMMETRIC | TAUCS_LOWER;*/
  PAPT->flags = A->flags;

  len    = (int*) malloc(n * sizeof(int));
  /*colptr = (int*) malloc(n * sizeof(int));*/
  if (!len) {
    taucs_printf("taucs_ccs_permute_symmetrically: out of memory\n");
    return NULL;
  }

  for (j=0; j<n; j++) len[j] = 0;

  for (j=0; j<n; j++) {
    for (ip = (A->colptr)[j]; ip < (A->colptr)[j+1]; ip++) {
      /*i = (A->rowind)[ip] - (A->indshift);*/
      i = (A->rowind)[ip];

      I = invperm[i];
      J = invperm[j];

      if (I < J) {
	int T = I; 
	I = J;
	J = T;
      }

      len[J] ++;
      
    }
  }

  (PAPT->colptr)[0] = 0;
  for (j=1; j<=n; j++) (PAPT->colptr)[j] = (PAPT->colptr)[j-1] + len[j-1];

  for (j=0; j<n; j++) len[j] = (PAPT->colptr)[j];
  
  for (j=0; j<n; j++) {
    for (ip = (A->colptr)[j]; ip < (A->colptr)[j+1]; ip++) {
      /*i   = (A->rowind)[ip] - (A->indshift);*/
      i   = (A->rowind)[ip];
      AIJ = (A->values)[ip];

      I = invperm[i];
      J = invperm[j];

      if (I < J) {
	int T = I; 
	I = J;
	J = T;
      }

      /*(PAPT->rowind)[ len[J] ] = I + (PAPT->indshift);*/
      (PAPT->rowind)[ len[J] ] = I;
      (PAPT->values)[ len[J] ] = AIJ;

      len[J] ++;
    }
  }

  free(len);
  return PAPT;
}

/* compute B=A*X, A must be square */

void 
taucs_ccs_times_vec(taucs_ccs_matrix* m, 
		    double* X,
		    double* B)
{
  int i,ip,j,n;
  double Aij;

  assert(m->flags & TAUCS_SYMMETRIC);
  assert(m->flags & TAUCS_LOWER);

  n = m->n;
  
  for (i=0; i < n; i++) B[i] = 0.0;

  for (j=0; j<n; j++) {
    for (ip = (m->colptr)[j]; ip < (m->colptr[j+1]); ip++) {
      i   = (m->rowind)[ip];
      Aij = (m->values)[ip];

      B[i] += X[j] * Aij;
      if (i != j) 
	B[j] += X[i] *Aij;
    }
  }
} 

taucs_ccs_matrix* 
taucs_ccs_augment_nonpositive_offdiagonals(taucs_ccs_matrix* A)
{
  int n,i,j;
  int *tmp;
  taucs_ccs_matrix* A_tmp;
  
  if (!(A->flags & TAUCS_SYMMETRIC) || !(A->flags & TAUCS_LOWER)) {
    taucs_printf("taucs_ccs_augment_nonpositive_offdiagonal: matrix not symmetric or not lower\n");
    return NULL;
  }

  n=A->n;

  tmp = (int *)calloc((2*n+1),sizeof(int));
  if (!tmp) {
    taucs_printf("taucs_ccs_augment_nonpositive_offdiagonal: out of memory\n");
    return NULL;
  }

  A_tmp = taucs_ccs_create(2*n,2*n,2*(A->colptr[n]));
  if (A_tmp == NULL) {
    free(tmp);
    return NULL;
  }
  A_tmp->flags |= TAUCS_SYMMETRIC | TAUCS_LOWER;
  
  
  for(i=0;i<n;i++) {
    for(j=A->colptr[i];j<A->colptr[i+1];j++) {
      if ((i == A->rowind[j])||(A->values[j] < 0)) {
	tmp[i]++;
	tmp[i+n]++;
      } else {
	tmp[i]++;
	tmp[A->rowind[j]]++;
      }
    }
  }

  A_tmp->colptr[0]=0;
  for(i=0;i<2*n;i++) A_tmp->colptr[i+1] = A_tmp->colptr[i] + tmp[i];
  for(i=0;i<2*n;i++) tmp[i] = A_tmp->colptr[i];
  
  for(i=0;i<n;i++) {
    for(j=A->colptr[i];j<A->colptr[i+1];j++) {
      if ((i == A->rowind[j])||(A->values[j] < 0)) {
	A_tmp->rowind[tmp[i]]=A->rowind[j];
	A_tmp->values[tmp[i]++]=A->values[j];
	A_tmp->rowind[tmp[i+n]]=A->rowind[j]+n;
	A_tmp->values[tmp[i+n]++]=A->values[j];
      } else {
	A_tmp->rowind[tmp[i]]=A->rowind[j]+n;
	A_tmp->values[tmp[i]++]=-A->values[j];
	A_tmp->rowind[tmp[A->rowind[j]]]=i+n;
	A_tmp->values[tmp[A->rowind[j]]++]=-A->values[j];
      }
    }
  }
  free(tmp);

  return A_tmp;
}

/*********************************************************/
/*                                                       */
/*********************************************************/

